﻿using System;
using System.IO;
using System.Text;

namespace Parade.Metadata.Asf
{
	/// <summary>
	/// ASF のタグ情報をダンプします。
	/// </summary>
	class TagDumper : IDisposable
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="filePath">ASF ファイルへのパス情報。</param>
		/// <param name="dumpFilePath">ダンプ先となるテキストファイルへのパス情報。</param>
		public TagDumper( string filePath, string dumpFilePath )
		{
			this._reader   = new FileStream( filePath, FileMode.Open, FileAccess.Read, FileShare.Read );
			this._writer   = new StreamWriter( dumpFilePath, true, Encoding.Unicode );
			this._filePath = filePath;
		}

		/// <summary>
		/// リソースを解放します。
		/// </summary>
		public void Dispose()
		{
			if( this._reader != null )
			{
				this._reader.Dispose();
				this._reader = null;
			}

			if( this._writer != null )
			{
				this._writer.Dispose();
				this._writer = null;
			}

			GC.SuppressFinalize( this );
		}

		/// <summary>
		/// ASF のタグ情報をダンプします。
		/// </summary>
		/// <exception cref="NotSupportedFileFormatException">ASF ファイルではありません。</exception>
		public void Dump()
		{
			var reader = this._reader;
			var writer = this._writer;

			// ASF ファイルであることをチェック
			reader.Seek( 0, SeekOrigin.Begin );
			{
				var info = new HeaderObjectInfo( reader );
				if( info.Guid != Guids.HeaderObject ) { throw new NotSupportedException( "ASF file format is not." ); }

				writer.WriteLine( "\n{0}\n\nHeader Object\nName\tSize\tValue\nPosition\t8\t0\nObject Size\t{1}", this._filePath, info.Size );
			}

			// オブジェクトの総数取得 & Reserved 1、2 のスキップ
			var count = reader.ReadInt32();
			reader.Seek( 2, SeekOrigin.Current );
			writer.WriteLine( "Object Count\t{0}\t{1}", 4, count );

			// オブジェクト読み取り
			for( var i = 0; i < count; ++i )
			{
				var top = reader.Position;
				var info = new HeaderObjectInfo( reader );

				if( info.Guid == Guids.StreamPropertiesObject )
				{
					writer.WriteLine( "\nStream Properties Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
					this.DumpStreamPropertiesObject( reader, writer );
				}
				else if( info.Guid == Guids.CodecListObject )
				{
					writer.WriteLine( "\nCodec List Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
					this.DumpCodecListObject( reader, writer );
				}
				else if( info.Guid == Guids.FilePropertiesObject )
				{
					writer.WriteLine( "\nFile Properties Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
					this.DumpFilePropertiesObject( reader, writer );
				}
				else if( info.Guid == Guids.ExtendedContentDescriptionObject )
				{
					writer.WriteLine( "\nExtended Content Description Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
					this.DumpExtendedContentDescriptionObject( reader, writer );
				}
				else if( info.Guid == Guids.HeaderExtensionObject )
				{
					writer.WriteLine( "\nHeader Extension Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
					this.DumpHeaderExtensionObject( reader, writer );
				}
				else if( info.Guid == Guids.ContentBrandingObject )
				{
					writer.WriteLine( "\nContent Branding Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
					this.DumpContentBrandingObject( reader, writer );
				}
				else if( info.Guid == Guids.ContentDescriptionObject )
				{
					writer.WriteLine( "\nContent Description Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
					this.DumpContentDescriptionObject( reader, writer );
				}
				else if( info.Guid == Guids.ContentEncryptionObject )
				{
					// DRMv1
					writer.WriteLine( "\nContent Encryption Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
				}
				else if( info.Guid == Guids.ExtendedContentEncryptionObject )
				{
					// DRMv7
					writer.WriteLine( "\nExtended Content Encryption Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
				}
				else if( info.Guid == Guids.StreamBitratePropertiesObject )
				{
					writer.WriteLine( "\nStream Bitrate Properties Object\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", top, info.Size );
					this.DumpStreamBitratePropertiesObject( reader, writer );
				}
				else
				{
					writer.WriteLine( "\n{0}\nName\tSize\tValue\nPosition\t8\t{1}\nObject Size\t8\t{2}", info.Guid, top, info.Size );
				}

				// 次のオブジェクト始点へ移動
				reader.Seek( top + info.Size, SeekOrigin.Begin );
			}

			// Header Object 以降のトップ レベル オブジェクト
			do
			{
				var top  = reader.Position;
				var info = new HeaderObjectInfo( reader );

				string name = null;
				if( info.Guid == Guids.DataObject )
				{
					name = "* Data Object";
				}
				else if( info.Guid == Guids.SimpleIndexObject )
				{
					name = "* Simple Index Object";
				}
				else if( info.Guid == Guids.IndexObject )
				{
					name = "Index Object";
				}
				else if( info.Guid == Guids.MediaObjectIndexObject )
				{
					name = "Media Object Index Object";
				}
				else if( info.Guid == Guids.TimecodeIndexObject )
				{
					name = "Timecode Index Object";
				}
				else
				{
					name = info.Guid.ToString();
				}

				// 次のオブジェクト始点へ移動
				writer.WriteLine( "\n{0}\nName\tSize\tValue\nPosition\t8\t{1}\nObject Size\t{2}", name, reader.Position - HeaderObjectInfo.ClassByteSize, info.Size );
				reader.Seek( top + info.Size, SeekOrigin.Begin );
			}
			while( reader.Position < reader.Length );
		}

		/// <summary>
		/// Codec List Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpCodecListObject( Stream reader, StreamWriter writer )
		{
			writer.WriteLine( "Reserved\t{0}\t{1}", 16, reader.ReadGuid() );

			var codecEntriesCount = reader.ReadInt32();
			writer.WriteLine( "Codec Entries Count\t{0}\t{1}", 4, codecEntriesCount );

			// Codec Name Length、Codec Description Length はバイト数ではなく文字数なので注意すること。
			// エンコードは UTF-16LE なので一文字は 2 byte となり、得られた文字数 x 2 が読み取るべきバイト数となる。
			// サロゲートペアの存在を考慮すると 4 byte もあり得るが、その場合は現在のデータ仕様では正しく読めない。
			// また、コーデック情報に ASCII 以外の文字が入ることは考えにくいので、2 byte として決めうちしておく。
			//
			for( var i = 0; i < codecEntriesCount; ++i )
			{
				writer.WriteLine( "[ {2} ] Type\t{0}\t{1}", 2, reader.ReadUInt16(), i );

				var length = reader.ReadUInt16();
				writer.WriteLine( "[ {2} ] Codec Name Length\t{0}\t{1}", 2, length, i );

				length *= 2;
				writer.WriteLine( "[ {2} ] Codec Name\t{0}\t{1}", length, reader.ReadStringUtf16( length ), i );

				length = reader.ReadUInt16();
				writer.WriteLine( "[ {2} ] Codec Description Length\t{0}\t{1}", 2, length, i );

				length *= 2;
				writer.WriteLine( "[ {2} ] Codec Description\t{0}\t{1}", length, reader.ReadStringUtf16( length ), i );

				length = reader.ReadUInt16();
				writer.WriteLine( "[ {2} ] Codec Infomation Length\t{0}\t{1}", 2, length, i );

				reader.Seek( length, SeekOrigin.Current );
			}
		}

		/// <summary>
		/// メタデータをダンプします。
		/// Extended Content Description Object、Metadata Object、Metadata Library Object の格納形式はほぼ共通なので、このメソッドで共通化します。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		/// <param name="number">メタデータの番号。</param>
		/// <param name="isMetadataObject">Metadata Object 系から利用する場合は true。それ以外は false。</param>
		private void DumpCommonMetadata( Stream reader, StreamWriter writer, int number, bool isMetadataObject )
		{
			var length = reader.ReadUInt16();
			writer.WriteLine( "[ {2} ] Name Length\t{0}\t{1}", 2, length, number );

			TagDataType type;
			if( isMetadataObject )
			{
				type = ( TagDataType )reader.ReadUInt16();
				writer.WriteLine( "[ {2} ] Data Type\t{0}\t{1}", 2, type, number );

				var dataLength = reader.ReadUInt32();
				writer.WriteLine( "[ {2} ] Data Length\t{0}\t{1}", 4, dataLength, number );

				writer.WriteLine( "[ {2} ] Name\t{0}\t{1}", length, reader.ReadStringUtf16( length ), number );
				length = ( ushort )dataLength;
			}
			else
			{
				writer.WriteLine( "[ {2} ] Name\t{0}\t{1}", length, reader.ReadStringUtf16( length ), number );

				type = ( TagDataType )reader.ReadUInt16();
				writer.WriteLine( "[ {2} ] Data Type\t{0}\t{1}", 2, type, number );

				length = reader.ReadUInt16();
				writer.WriteLine( "[ {2} ] Data Length\t{0}\t{1}", 2, length, number );
			}

			switch( type )
			{
			case TagDataType.Bool: writer.WriteLine( "[ {2} ] Data\t{0}\t{1}", length, reader.ReadBool( length ), number ); break;
			case TagDataType.String: writer.WriteLine( "[ {2} ] Data\t{0}\t{1}", length, reader.ReadStringUtf16( length ), number ); break;
			case TagDataType.UInt16: writer.WriteLine( "[ {2} ] Data\t{0}\t{1}", length, reader.ReadUInt16( length ), number ); break;
			case TagDataType.UInt32: writer.WriteLine( "[ {2} ] Data\t{0}\t{1}", length, reader.ReadUInt32( length ), number ); break;
			case TagDataType.UInt64: writer.WriteLine( "[ {2} ] Data\t{0}\t{1}", length, reader.ReadUInt64( length ), number ); break;

			// バイト配列、あるいは未知のデータ型
			default:
				reader.Seek( length, SeekOrigin.Current );
				writer.WriteLine( "[ {2} ] Data\t{0}\t{1}", length, "Byte array", number );
				break;
			}
		}

		/// <summary>
		/// Content Branding Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpContentBrandingObject( Stream reader, StreamWriter writer )
		{
			writer.WriteLine( "Banner Image Type\t{0}\t{1}", 4, reader.ReadUInt32() );

			var length = reader.ReadInt32();
			writer.WriteLine( "Banner Image Data Size\t{0}\t{1}", 4, length );

			reader.Seek( length, SeekOrigin.Current );
			writer.WriteLine( "Banner Image Data\t{0}\t{1}", length, "Byte array..." );

			length = reader.ReadInt32();
			writer.WriteLine( "Banner Image URL Length\t{0}\t{1}", 4,      length                           );
			writer.WriteLine( "Banner Image URL\t{0}\t{1}",        length, reader.ReadStringAscii( length ) );

			length = reader.ReadInt32();
			writer.WriteLine( "Copyright URL Length\t{0}\t{1}", 4,      length                           );
			writer.WriteLine( "Copyright URL\t{0}\t{1}",        length, reader.ReadStringAscii( length ) );
		}

		/// <summary>
		/// Content Description Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpContentDescriptionObject( Stream reader, StreamWriter writer )
		{
			var lengths = new int[ 5 ] { reader.ReadInt16(), reader.ReadInt16(), reader.ReadInt16(), reader.ReadInt16(), reader.ReadInt16() };
			writer.WriteLine( "Title Length\t{0}\t{1}",       2, lengths[ 0 ] );
			writer.WriteLine( "Author Length\t{0}\t{1}",      2, lengths[ 1 ] );
			writer.WriteLine( "Copyright Length\t{0}\t{1}",   2, lengths[ 2 ] );
			writer.WriteLine( "Description Length\t{0}\t{1}", 2, lengths[ 3 ] );
			writer.WriteLine( "Rating Length\t{0}\t{1}",      2, lengths[ 4 ] );

			writer.WriteLine( "Title\t{0}\t{1}",       lengths[ 0 ], reader.ReadStringUtf16( lengths[ 0 ] ) );
			writer.WriteLine( "Author\t{0}\t{1}",      lengths[ 1 ], reader.ReadStringUtf16( lengths[ 1 ] ) );
			writer.WriteLine( "Copyright\t{0}\t{1}",   lengths[ 2 ], reader.ReadStringUtf16( lengths[ 2 ] ) );
			writer.WriteLine( "Description\t{0}\t{1}", lengths[ 3 ], reader.ReadStringUtf16( lengths[ 3 ] ) );
			writer.WriteLine( "Rating\t{0}\t{1}",      lengths[ 4 ], reader.ReadStringUtf16( lengths[ 4 ] ) );
		}

		/// <summary>
		/// Extended Content Description Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpExtendedContentDescriptionObject( Stream reader, StreamWriter writer )
		{
			ushort count = reader.ReadUInt16();
			writer.WriteLine( "Content Descriptors Count\t{0}\t{1}", 2, count );

			for( ushort i = 0; i < count; ++i )
			{
				this.DumpCommonMetadata( reader, writer, i, false );
			}
		}

		/// <summary>
		/// Header Extension Object 内における Extended Stream Properties Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpExtendedStreamPropertiesObject( Stream reader, StreamWriter writer )
		{
			writer.WriteLine( "Start Time\t{0}\t{1}",                        8, reader.ReadUInt64() );
			writer.WriteLine( "End Time\t{0}\t{1}",                          8, reader.ReadUInt64() );
			writer.WriteLine( "Data Bitrate\t{0}\t{1}",                      4, reader.ReadUInt32() );
			writer.WriteLine( "Buffer Size\t{0}\t{1}",                       4, reader.ReadUInt32() );
			writer.WriteLine( "Initial Buffer Fullness\t{0}\t{1}",           4, reader.ReadUInt32() );
			writer.WriteLine( "Alternate Data Bitrate\t{0}\t{1}",            4, reader.ReadUInt32() );
			writer.WriteLine( "Alternate Buffer Size\t{0}\t{1}",             4, reader.ReadUInt32() );
			writer.WriteLine( "Alternate Initial Buffer Fullness\t{0}\t{1}", 4, reader.ReadUInt32() );
			writer.WriteLine( "Maximum Object Size\t{0}\t{1}",               4, reader.ReadUInt32() );
			writer.WriteLine( "Flags\t{0}\t0x{1:X8}",                        4, reader.ReadUInt32() );
			writer.WriteLine( "Stream Number\t{0}\t{1}",                     2, reader.ReadUInt16() );
			writer.WriteLine( "Stream Language ID Index\t{0}\t{1}",          2, reader.ReadUInt16() );
			writer.WriteLine( "Average Time Per Frame\t{0}\t{1}",            8, reader.ReadUInt64() );

			ushort streamNamesCount = reader.ReadUInt16();
			writer.WriteLine( "Stream Name Count\t{0}\t{1}", 2, streamNamesCount );

			ushort payloadExtensionSystemCount = reader.ReadUInt16();
			writer.WriteLine( "Payload Extension System Count\t{0}\t{1}", 2, payloadExtensionSystemCount );

			// Stream Names
			for( ushort i = 0; i < streamNamesCount; ++i )
			{
				writer.WriteLine( "[ {2} ] Language ID Index\t{0}\t{1}", 2, reader.ReadUInt16(), i );

				var length = reader.ReadUInt16();
				writer.WriteLine( "[ {2} ] Stream Name Length\t{0}\t{1}", 2,      length,                           i );
				writer.WriteLine( "[ {2} ] Stream Name\t{0}\t{1}",        length, reader.ReadStringUtf16( length ), i );
			}

			// Payload Extension Systems
			for( ushort i = 0; i < payloadExtensionSystemCount; ++i )
			{
				writer.WriteLine( "[ {2} ] Extension System ID\t{0}\t{1}", 16, reader.ReadGuid(),   i );
				writer.WriteLine( "[ {2} ] Extension Data Size\t{0}\t{1}",  2, reader.ReadUInt16(), i );

				var length = reader.ReadUInt32();
				writer.WriteLine( "[ {2} ] Extension System Info Length\t{0}\t{1}", 4, length, i );

				reader.Seek( length, SeekOrigin.Current );
				writer.WriteLine( "[ {2} ] Stream Name\t{0}\t{1}", length, "Byte Array ...", i );
			}

			// ... Stream Properties Object
		}

		/// <summary>
		/// File Properties Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpFilePropertiesObject( Stream reader, StreamWriter writer )
		{
			writer.WriteLine( "File ID\t{0}\t{1}",                  16, reader.ReadGuid()   );
			writer.WriteLine( "File Size\t{0}\t{1}",                 8, reader.ReadUInt64() );
			writer.WriteLine( "Creation Date\t{0}\t{1}",             8, reader.ReadUInt64() );
			writer.WriteLine( "Data Packets Count\t{0}\t{1}",        8, reader.ReadUInt64() );
			writer.WriteLine( "Play Duration\t{0}\t{1}",             8, reader.ReadUInt64() );
			writer.WriteLine( "Send Duration\t{0}\t{1}",             8, reader.ReadUInt64() );
			writer.WriteLine( "Preroll\t{0}\t{1}",                   8, reader.ReadUInt64() );
			writer.WriteLine( "Flags\t{0}\t0x{1:X4}",                4, reader.ReadUInt32() );
			writer.WriteLine( "Minimum Data Packet Size\t{0}\t{1}",  4, reader.ReadUInt32() );
			writer.WriteLine( "Maximum Data Packet Size\t{0}\t{1}",  4, reader.ReadUInt32() );
			writer.WriteLine( "Maximum Bitrate\t{0}\t{1}",           4, reader.ReadUInt32() );
		}

		/// <summary>
		/// Header Extension Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpHeaderExtensionObject( Stream reader, StreamWriter writer )
		{
			writer.WriteLine( "Reserved 1\t{0}\t{1}", 16, reader.ReadGuid() );
			writer.WriteLine( "Reserved 2\t{0}\t{1}", 2, reader.ReadUInt16() );

			var size = reader.ReadInt32();
			writer.WriteLine( "Header Extension Data Size\t{0}\t{1}", 4, size );

			var end = reader.Position + size;
			do
			{
				var begin = reader.Position;
				var info = new HeaderObjectInfo( reader );

				if( info.Guid == Guids.MetadataObject )
				{
					writer.WriteLine( "\nHeader Extension Object [ Metadata Object ]\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", reader.Position - HeaderObjectInfo.ClassByteSize, info.Size );
					this.DumpMetadataObject( reader, writer );
				}
				else if( info.Guid == Guids.MetadataLibraryObject )
				{
					writer.WriteLine( "\nHeader Extension Object [ Metadata Library Object ]\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", reader.Position - HeaderObjectInfo.ClassByteSize, info.Size );
					this.DumpMetadataLibraryObject( reader, writer );
				}
				else if( info.Guid == Guids.LanguageListObject )
				{
					writer.WriteLine( "\nHeader Extension Object [ Language List Object ]\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", reader.Position - HeaderObjectInfo.ClassByteSize, info.Size );
					this.DumpLanguageListObject( reader, writer );
				}
				else if( info.Guid == Guids.PaddingObject )
				{
					writer.WriteLine( "\nHeader Extension Object [ Padding Object ]\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", reader.Position - HeaderObjectInfo.ClassByteSize, info.Size );
				}
				else if( info.Guid == Guids.StreamPrioritizationObject )
				{
					writer.WriteLine( "\nHeader Extension Object [ Stream Prioritization Object ]\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", reader.Position - HeaderObjectInfo.ClassByteSize, info.Size );
					this.DumpStreamPrioritizationObject( reader, writer );
				}
				else if( info.Guid == Guids.ExtendedStreamPropertiesObject )
				{
					writer.WriteLine( "\nHeader Extension Object [ Extended Stream Properties Object ]\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", reader.Position - HeaderObjectInfo.ClassByteSize, info.Size );
					this.DumpExtendedStreamPropertiesObject( reader, writer );
				}
				else
				{
					writer.WriteLine( "\nHeader Extension Object [ {2} ]\nName\tSize\tValue\nPosition\t8\t{0}\nObject Size\t8\t{1}", reader.Position - HeaderObjectInfo.ClassByteSize, info.Size, info.Guid );
				}

				reader.Seek( begin + info.Size, SeekOrigin.Begin );
			}
			while( end > reader.Position );
		}

		/// <summary>
		/// Header Extension Object 内における Language List Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpLanguageListObject( Stream reader, StreamWriter writer )
		{
			ushort count = reader.ReadUInt16();
			writer.WriteLine( "Language ID Records Count\t{0}\t{1}", 2, count );

			for( ushort i = 0; i < count; ++i )
			{
				var length = reader.ReadByte();
				writer.WriteLine( "[ {2} ] Language ID Length\t{0}\t{1}", 1, length,                           i );
				writer.WriteLine( "[ {2} ] Language ID\t{0}\t{1}",        2, reader.ReadStringUtf16( length ), i );
			}
		}

		/// <summary>
		/// Header Extension Object 内における Metadata Library Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpMetadataLibraryObject( Stream reader, StreamWriter writer )
		{
			ushort count = reader.ReadUInt16();
			writer.WriteLine( "Description Records Count\t{0}\t{1}", 2, count );

			for( ushort i = 0; i < count; ++i )
			{
				writer.WriteLine( "[ {2} ] Language List Index\t{0}\t{1}", 2, reader.ReadUInt16(), i );
				writer.WriteLine( "[ {2} ] Stream Number\t{0}\t{1}",       2, reader.ReadUInt16(), i );
				this.DumpCommonMetadata( reader, writer, i, true );
			}
		}

		/// <summary>
		/// Header Extension Object 内における Metadata Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpMetadataObject( Stream reader, StreamWriter writer )
		{
			ushort count = reader.ReadUInt16();
			writer.WriteLine( "Description Records Count\t{0}\t{1}", 2, count );

			for( ushort i = 0; i < count; ++i )
			{
				writer.WriteLine( "[ {2} ] Reserved\t{0}\t{1}",      2, reader.ReadUInt16(), i );
				writer.WriteLine( "[ {2} ] Stream Number\t{0}\t{1}", 2, reader.ReadUInt16(), i );
				this.DumpCommonMetadata( reader, writer, i, true );
			}
		}

		/// <summary>
		/// Stream Bitrate Properties Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpStreamBitratePropertiesObject( Stream reader, StreamWriter writer )
		{
			var count = reader.ReadUInt16();
			writer.WriteLine( "Bitrate Records Count\t{0}\t{1}", 2, count );

			for( ushort i = 0; i < count; ++i )
			{
				writer.WriteLine( "Flags\t{0}\t0x{1:X4}",      2, reader.ReadUInt16() );
				writer.WriteLine( "Average Bitrate\t{0}\t{1}", 4, reader.ReadUInt32() );
			}
		}

		/// <summary>
		/// Header Extension Object 内における Stream Prioritization Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpStreamPrioritizationObject( Stream reader, StreamWriter writer )
		{
			ushort count = reader.ReadUInt16();
			writer.WriteLine( "Priority Records Count\t{0}\t{1}", 2, count );

			for( ushort i = 0; i < count; ++i )
			{
				writer.WriteLine( "[ {2} ] Stream Number\t{0}\t{1}",       2, reader.ReadUInt16(), i );
				writer.WriteLine( "[ {2} ] Priority Flags\t{0}\t0x{1:X4}", 2, reader.ReadUInt16(), i );
			}
		}

		/// <summary>
		/// Stream Properties Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void DumpStreamPropertiesObject( Stream reader, StreamWriter writer )
		{
			var streamType = reader.ReadGuid();
			writer.WriteLine( "Stream Type\t{0}\t{1}",                 16, streamType          );
			writer.WriteLine( "Error Correction Type\t{0}\t{1}",       16, reader.ReadGuid()   );
			writer.WriteLine( "Time Offset\t{0}\t{1}",                  8, reader.ReadUInt64() );
			writer.WriteLine( "Type-Specific Data Length\t{0}\t{1}",    4, reader.ReadUInt32() );
			writer.WriteLine( "Erro Correction Data Length\t{0}\t{1}",  4, reader.ReadUInt32() );
			writer.WriteLine( "Flags\t{0}\t0x{1:X4}",                   2, reader.ReadUInt16() );
			writer.WriteLine( "Reserved\t{0}\t{1}",                     4, reader.ReadUInt32() );

			if( streamType == Guids.VideoMedia )
			{
				writer.WriteLine( "Encoded Image Width\t{0}\t{1}",  4, reader.ReadUInt32() );
				writer.WriteLine( "Encoded Image Height\t{0}\t{1}", 4, reader.ReadUInt32() );
				writer.WriteLine( "Reserved Flags\t{0}\t0x{1:X4}",  1, reader.ReadByte()   );

				var formatDataSize = reader.ReadUInt16();
				writer.WriteLine( "Format Data Size\t{0}\t{1}", 2, formatDataSize );

				if( formatDataSize > 0 )
				{
					var bmpInfo = new BitmapInfoHeader( reader );
					writer.WriteLine( "BMP - FormatDataSize\t{0}\t{1}",              4, bmpInfo.FormatDataSize           );
					writer.WriteLine( "BMP - Image Width\t{0}\t{1}",                 4, bmpInfo.ImageWidth               );
					writer.WriteLine( "BMP - Image Height\t{0}\t{1}",                4, bmpInfo.ImageHeight              );
					writer.WriteLine( "BMP - Reserved\t{0}\t{1}",                    2, bmpInfo.Reserved                 );
					writer.WriteLine( "BMP - Bit Per Pixcel Count\t{0}\t{1}",        2, bmpInfo.BitPerPixcelCount        );
					writer.WriteLine( "BMP - Compression ID\t{0}\t0x{1:X8}",         4, bmpInfo.CompressionId            );
					writer.WriteLine( "BMP - Image Size\t{0}\t{1}",                  4, bmpInfo.ImageSize                );
					writer.WriteLine( "BMP - Horizontal Pixcel Per Meter\t{0}\t{1}", 4, bmpInfo.HorizontalPixcelPerMeter );
					writer.WriteLine( "BMP - Vertical Pixcel Per Meter\t{0}\t{1}",   4, bmpInfo.VerticalPixcelPerMeter   );
					writer.WriteLine( "BMP - Color Used Count\t{0}\t{1}",            4, bmpInfo.ColorUsedCount           );
					writer.WriteLine( "BMP - Important Color Count\t{0}\t{1}",       4, bmpInfo.ImportantColorCount      );
				}
			}
		}

		/// <summary>
		/// ASF ファイルのパス情報。
		/// </summary>
		private string _filePath;

		/// <summary>
		/// ASF のタグ情報を読み取るためのストリーム。
		/// </summary>
		private Stream _reader;

		/// <summary>
		/// タグ情報を出力するためのストリーム。
		/// </summary>
		private StreamWriter _writer;
	}
}
